#!/usr/bin/perl -w
#########################################################################################
#
# Hex-to-SysEx converter
# Original Author: tk@midibox.org (2003-01-30) (for PIC version)
# Author: green@commonsound.com (2009-04-08) (converted to AVR)
#
# Converts a INHX32 file to a SysEx file to be uploaded to an AVR using MIDI-OX
#
#########################################################################################
#
# F0		Start Sysex
# 04 33 55	Vendor: 4ms
# 01		Bootload data command
# 00 01		Length of page when unpacked, 128 bytes default  (128 == 00 01 in 7-bit packed format)
# AL AH		Address: 00 00, 40 00, 00 01, 40 01, 00 02, 40 02,...
# 00 00		upper address bytes (not used)
# <dump>	Data: (147 7-bit bytes unpacks to 128 8-bit bytes)
# XX		Checksum byte
# F7		End Sysex
#########################################################################################
# Then, firmware checksum data is sent:
#
# F0		Start Sysex
# 04 33 55	Vendor: 4ms
# 03		Write checksum command
# XX XX XX	Length of firmware (actual code)
# XX XX		Checksum calculated by XOR'ing each (unpacked) byte of the actual code
# F7		End Sysex
#########################################################################################


my $vendor_1=0x04;
my $vendor_2=0x33;
my $vendor_3=0x55;

$pagesize=128;

use Getopt::Long;


GetOptions (
   "pagesize=s" => \$pagesize,
   "debug" => \$debug,
);

if( scalar(@ARGV) != 1 )
{
   die "SYNTAX:   avrhex2syx.pl <hex-file>.hex [<-pagesize SIZE>] [<-debug>]\n";
}


$in_file  = $ARGV[0];

if( lc(substr($in_file, length($in_file)-4)) ne ".hex" )
{
   die "ERROR: File doesn't end with '.hex'\n";
}

$out_file = substr($in_file, 0, -4) . ".syx";

open(IN, "<${in_file}") || die "ERROR: Cannot open ${in_file} for reading!\n";




####################################################################################
# Read in .hex file
####################################################################################

my $line = 0;
my %dump = ();

my $end_read = 0;
my $address_extension = 0;

my $firmware_checksum = 0;
my $firmware_length = 0;

while( <IN> )
{
   ++$line;
   s/[\n\r]//g;

   if( !$end_read && substr($_, 0, 1) eq ":" && length($_) >= 11 && length($_)%2 && !/[^:0-9a-f]/i )
   {
      my @record = ();
      my $i;

      for($i=1; $i<length($_); $i+=2)
      {
         push @record, hex substr($_, $i, 2);
      }

      my $number_bytes = $record[0];
      my $address      = ($record[1] << 8) | $record[2];
      my $record_type  = $record[3];
      my $checksum     = $record[-1];

      if( scalar(@record) != ($number_bytes+5) )
      {
         die "ERROR: wrong number of bytes in line ${line}\n";
      }

      my $calculated_checksum = 0x00;
      for($i=0; $i<scalar(@record)-1; ++$i)
      {
         $calculated_checksum = ($calculated_checksum + $record[$i]) & 0xff;
      }
      $calculated_checksum = (-$calculated_checksum) & 0xff;

      if( $debug )
      {
         printf("%02X %04X %02X ... %02X (%02X)\n", $number_bytes, $address, $record_type, $checksum, $calculated_checksum);
      }

      if( $calculated_checksum != $checksum )
      {
         die "ERROR: wrong checksum in line ${line}\n";
      }


      if   ( $record_type == 0x00 )
      {
         for($i=0; $i<$number_bytes; ++$i)
         {
				my $a = $address_extension + $address + $i;

            if( exists $dump{$a} ) {die sprintf "ERROR: in line ${line} - address %X already allocated\n", $a;}

            $dump{$a} = $record[4+$i];
				$firmware_checksum += $record[4+$i];
				$firmware_length++;
	
            if( $debug ) {printf "[%08X] %02X - Firmware Checksum running tally: %06X\n", $a, $dump{$a}, $firmware_checksum;}

      	}
      }
      elsif( $record_type == 0x01 )
      {
         $end_read = 1;
      }
      elsif( $record_type == 0x04 )
      {
         if( scalar(@record) != 5+2 )
         {
            die "ERROR: for an INHX32 record (0x04) expecting 2 address bytes!\n";
         }
         $address_extension = (($record[4] << 8) | $record[5]) << 16;
			print "LOG: line {$line} contains an address_extension of {$address_extension}, (record_type=0x04)";
      }
      else
      {
         die "ERROR: unknown record type in line ${line}\n";
      }
   }
   elsif( $end_read )
   {
      print "WARNING: line {$line} ignored, since END record has been notified\n";
   }
   elsif( length($_) )
   {
      print "WARNING: invalid line ${line}: '$_'\n";
   }
   else
   {
      print "WARNING: line ${line} is empty.\n";
   }
}		

close(IN);

$firmware_checksum &= 0x3FFF;

#########################################################################################
# Write out .syx file
#########################################################################################

open(OUT, ">${out_file}") || die "ERROR: Cannot open ${out_file} for writing!\n";
binmode(OUT);


for($address=0; $address<0xffff; $address+=$pagesize)
{

   my $block_allocated = 0;
   my @out_dump = ();

   my $offset;
   for($offset=0; $offset<$pagesize; ++$offset)
   {
      my $a = $address + $offset;
      my $b = 0x00;

      if( exists $dump{$a} )
      {
         $block_allocated = 1;
         $b = $dump{$a};
      }
      else
      {
         $b = 0xff;
      }
      $out_dump[$offset] = $b;
   }


   if( $block_allocated )
   {
      my $checksum = 0;
      sysex_write_header($address, $pagesize, \$checksum);


		my $cnt7 = 0;
		my @bitblock = ();
		my $b;
		my $msb;
		my $blockstart=0;

		for ($offset = 0; $offset < $pagesize; $offset++){
			$b = $out_dump[$offset] & 0x7F;
			$msb = $out_dump[$offset] >> 7;

			$bitblock[$blockstart] |= $msb << $cnt7;
			$bitblock[$blockstart + 1 + $cnt7] = $b; 	

			$cnt7++;
			if ($cnt7 == 7){
					$blockstart += 8;
					$bitblock[$blockstart]=0;
					$cnt7 = 0;
			}
		}

		$blockstart += $cnt7 + ($cnt7 !=0 ? 1 : 0);
		
		for (my $j=0;$j<$blockstart;$j++) {
			print OUT sprintf("%c", $bitblock[$j]);
			$checksum ^=  $bitblock[$j];
		}

		$checksum &= 0x7F;

		sysex_write_footer($checksum);

		printf("Block %04X-%04X allocated - Checksum: %02X\n", $address, $address+$pagesize-1, $checksum);
	
   }
}


#########################################################################################
# Print firmware checksum write command
# This tells the receiving AVR to burn the checksum data into the EEPROM,
# which is used by the bootstrap to check for damage to code memory on bootup 
#########################################################################################

my $chkl = $firmware_checksum & 0x7F;
my $chkh = ($firmware_checksum >> 7) & 0x7F;

my $lenl = $firmware_length & 0x7F;
my $lenh = ($firmware_length >> 7) & 0x7F;
my $lenm = ($firmware_length >> 14) & 0x7F;

print OUT sprintf "%c%c%c%c%c%c%c%c%c%c%c",
      0xf0, $vendor_1, $vendor_2, $vendor_3, 0x03, $lenl, $lenh, $lenm, $chkl, $chkh, 0xf7;

close(OUT);

printf("Firmware length: %06X.\nFirmware checksum: %04X\n", $firmware_length, $firmware_checksum);

printf("Vendor ID: %02X %02X %02X\n", $vendor_1, $vendor_2, $vendor_3);



#########################################################################################
# Header/Footer subs
#########################################################################################

sub sysex_write_header
{
   my ($address, $range, $checksum) = @_;

   my $cmd = 0x01;

   my $am = ($address >> 14) & 0x7f;
   my $ah = ($address >> 7) & 0x7f;
   my $al = $address & 0x7f;

   my $rh = ($range >> 7) & 0x7f;
   my $rl = $range & 0x7f;

   print OUT sprintf "%c%c%c%c%c%c%c%c%c%c",
      0xf0, $vendor_1, $vendor_2, $vendor_3, $cmd, $rl, $rh, $al, $ah, $am;

   $$checksum =  $cmd ^ $rl ^ $rh ^ $al ^ $ah ^ $am;
}

sub sysex_write_footer
{
   my ($checksum) = @_;

   print OUT sprintf "%c%c", $checksum, 0xf7;
}
